import {fetchAllFeedback, fetchAllAssignments} from '@/api'
import {objectifyArray} from '@/util/helpers'
import {Assignment, Feedback, Subscription} from '@/models'
import {Module} from 'vuex'
import {RootState} from '@/store/store'

export const feedbackTableMut = Object.freeze({
  SET_FEEDBACK_HISTORY: 'SET_FEEDBACK_HISTORY',
  ADD_ASSIGNMENTS_INFO: 'ADD_ASSIGNMENTS_INFO',
  SET_FEEDBACK_OF_SUBMISSION_TYPE: 'SET_FEEDBACK_OF_SUBMISSION_TYPE',
  MAP_FEEDBACK_HIST_OF_SUBMISSION_TYPE: 'MAP_FEEDBACK_HIST_OF_SUBMISSION_TYPE',
  RESET_STATE: 'RESET_STATE'
})

export interface FeedbackHistoryItem extends Feedback {
  history?: {
    [key in Subscription.FeedbackStageEnum]: {
      owner: string
      isDone: boolean
    }
  }
}

interface FeedbackTableState {
    feedbackHist: {[submissionPk: string]: FeedbackHistoryItem}
}

function initialState (): FeedbackTableState {
  return {
    feedbackHist: {}
  }
}

const feedbackTable: Module<FeedbackTableState, RootState> = {
  state: initialState(),
  mutations: {
    [feedbackTableMut.SET_FEEDBACK_HISTORY]: (state, val: Array<Feedback>) => {
      state.feedbackHist = objectifyArray(val, 'ofSubmission')
    },
    [feedbackTableMut.ADD_ASSIGNMENTS_INFO]: (state, assignments: Array<Assignment>) => {
      for (const assignment of assignments) {
        if (!assignment.submission || !assignment.stage) {
          throw Error()
        }
        const feedback = state.feedbackHist[assignment.submission]
        feedback.history = {
          ...feedback.history,
          [assignment.stage]: {
            owner: assignment.owner,
            isDone: assignment.isDone
          }
        }
      }
    },
    [feedbackTableMut.SET_FEEDBACK_OF_SUBMISSION_TYPE] (state, {feedback, type}) {
      state.feedbackHist[feedback.ofSubmission].ofSubmissionType = type.name
    },
    [feedbackTableMut.RESET_STATE]: (state) => { Object.assign(state, initialState()) }
  },
  actions: {
    mapFeedbackHistOfSubmissionType ({getters, commit, state}) {
      for (const feedback of Object.values(state.feedbackHist)) {
        const type = getters.getSubmissionType((feedback as any).ofSubmissionType)
        commit(feedbackTableMut.SET_FEEDBACK_OF_SUBMISSION_TYPE, {feedback, type})
      }
    },
    async getFeedbackHistory ({getters, commit, dispatch}) {
      let data: [Promise<Feedback[]>, Promise<Assignment[]> | undefined] =
          [fetchAllFeedback(), getters.isReviewer ? fetchAllAssignments() : undefined]

      Promise.all<Feedback[], Assignment[] | undefined>(data)
        .then(([feedbacks, assignments]: [Feedback[], Assignment[]?]) => {
          commit(feedbackTableMut.SET_FEEDBACK_HISTORY, feedbacks)
          dispatch('mapFeedbackHistOfSubmissionType')
          if (assignments) {
            commit(feedbackTableMut.ADD_ASSIGNMENTS_INFO, assignments)
          }
        })
    }
  }
}

export default feedbackTable
